/* File:      parm.C
   Date:      6/22/02 (revised from fall 2001 version)
   Author:    C. Paciorek
   Purpose:   code for parameter objects, including MCMC proposal, evaluation and updating
   Compile:   g++ -c 
   Usage:     
   Arguments: 
   Requires:  base.h,parm.h,cov_mat.h,transform.h,density.h,ran1.h,gasdev.h,dataset.h,locations.h
   Reference: this code is somewhat similar to the spatial process parameterization of 
              Higdon 1998, Env Ecol Statistics describes the discretization and parameterization used here, 
              which bears similarity to the non-discretized version of Lockwood 2001, PhD thesis from CMU 
              Statistics, which describes a reparameterization used for spatially correlated random variables
*/

#include "base.h"
#include "parm.h"
#include "density.h"
#include "gasdev.h"
#include "ran1.h"
#include "cov_mat.h"
#include "dataset.h"
#include "transform.h"
#include "locations.h"

///////////////////////////////////////////////////////
// Class nonspatial_parm
///////////////////////////////////////////////////////

nonspatial_parm::nonspatial_parm(int which_in,double theta_init,double theta_mean, double theta_sd, double prop_sd_in,prior_dens_fxn_ptr prior_in,transform_fxn_ptr transform_in){
  int accepted=ACCEPT;
  which=which_in;            // which parameter is it
  psi=new hyperparm_scal(theta_init,theta_mean,theta_sd,prop_sd_in,prior_in,transform_in);  
  psi->evaluate_prior();     // initialize prior density
  psi->update(accepted);     // put same value in stored 
  num_accept=0;
  run_num_accept=0;
}

nonspatial_parm::~nonspatial_parm(void){
  delete psi;
}

int nonspatial_parm::accept(double logprior_diff){
  float unif;
  unif=ran1(idum,GENERATE,"a"); 
  double tmp=exp(data->evaluate_loglik_diff(which,NOT_FIRST));
  if(tmp*exp(logprior_diff)>unif){ 
    return(ACCEPT);
  }
  else{
    return(REJECT);
  }
}


int nonspatial_parm::step(void){
  int accepted=REJECT;
  double logprior_diff;

  if(!psi->fixed){
    psi->propose(0.0);
    logprior_diff=psi->evaluate_prior();
    if(zero(logprior_diff-NEG_INFTY)){
      psi->update(accepted);
   }
    else{
      data->datacov_change(which,NOT_FIRST);
      accepted=accept(logprior_diff);
      num_accept+=accepted;
      run_num_accept+=accepted;
      psi->update(accepted);
      data->update(accepted,which);  // update likelihood
    }
  }
  return(accepted);
}

///////////////////////////////////////////////////////
// Class nonspatial_vecparm
///////////////////////////////////////////////////////

// this is used for being able to jointly propose parameters concerning basis kernels so as to reduce computation

nonspatial_vecparm::nonspatial_vecparm(int which_in,int num_in,double* theta_init,double* theta_mean, double* theta_sd, double prop_sd_in,prior_dens_fxn_ptr prior_in,transform_fxn_ptr transform_in){
  int accepted=ACCEPT;
  num=num_in;
  which=which_in;            // which parameter is it
  psi=new (hyperparm_scal*)[num];
  ptrvec=new (double*)[num];
  for(int i=0;i<num;i++){
    psi[i]=new hyperparm_scal(theta_init[i],theta_mean[i],theta_sd[i],prop_sd_in,prior_in,transform_in);
    ptrvec[i]=&(psi[i]->value);
    if(which==DATA_COVAR_EIGVEC){
      while(psi[i]->base_value<0){
        psi[i]->base_value+=PI;
      }
      while(psi[i]->base_value>PI){
        psi[i]->base_value-=PI;
      }
      psi[i]->value=psi[i]->base_value;
    }
    psi[i]->evaluate_prior();  // initialize prior density
    psi[i]->update(accepted);  // put same value in store
  }
  num_accept=0;
  run_num_accept=0;
}

nonspatial_vecparm::nonspatial_vecparm(int which_in,int num_in,double* theta_init,double theta_mean, double theta_sd, double prop_sd_in,prior_dens_fxn_ptr prior_in,transform_fxn_ptr transform_in){
  int accepted=ACCEPT;
  num=num_in;
  which=which_in;            // which parameter is it
  psi=new (hyperparm_scal*)[num];
  ptrvec=new (double*)[num];
  for(int i=0;i<num;i++){
    psi[i]=new hyperparm_scal(theta_init[i],theta_mean,theta_sd,prop_sd_in,prior_in,transform_in);
    ptrvec[i]=&(psi[i]->value);
    if(which==DATA_COVAR_EIGVEC){
      while(psi[i]->base_value<0){
        psi[i]->base_value+=PI;
      }
      while(psi[i]->base_value>PI){
        psi[i]->base_value-=PI;
      }
      psi[i]->value=psi[i]->base_value;
    }
    psi[i]->evaluate_prior();  // initialize prior density
    psi[i]->update(accepted);  // put same value in store
  }
  num_accept=0;
  run_num_accept=0;
}

nonspatial_vecparm::~nonspatial_vecparm(void){
  for(int i=0;i<num;i++){  
    delete psi[i];
  }
  delete [] psi;
  delete [] ptrvec;
}

int nonspatial_vecparm::accept(double logprior_diff){
  float unif;
  unif=ran1(idum,GENERATE,"a"); 
  double tmp=exp(data->evaluate_loglik_diff(which,NOT_FIRST));
  if(tmp*exp(logprior_diff)>unif){ 
    return(ACCEPT);
  }
  else{
    return(REJECT);
  }
}

int nonspatial_vecparm::step(int mode){
  int accepted=REJECT;
  double logprior_diff,tmp;
  int i;
  int count_rej=0;

  if(!psi[0]->fixed){  // this depends on input prop_sd, so if one is fixed, all are
    logprior_diff=0.0;
    for(i=0;i<num;i++){
      psi[i]->propose(0.0);
      if(which==DATA_COVAR_EIGVEC){
        while(psi[i]->base_value<0){
          psi[i]->base_value+=PI;
        }
        while(psi[i]->base_value>PI){
          psi[i]->base_value-=PI;
        }
        psi[i]->value=psi[i]->base_value;
      }
      tmp=psi[i]->evaluate_prior();
      if(zero(tmp-NEG_INFTY)){
        count_rej++;
      }
      logprior_diff+=tmp;
    }
    if(count_rej){
      for(i=0;i<num;i++){
        psi[i]->update(accepted);
      }
    }
    else{
      data->datacov_change(which,NOT_FIRST);
      if(mode==POSTMEAN){
        logprior_diff+=alpha->ksi_propose(FALSE);
        logprior_diff+=alpha->theta->evaluate_prior();
        logprior_diff+=beta->ksi_propose(FALSE);
        logprior_diff+=beta->theta->evaluate_prior();
        alpha->ksi_to_psi();
        beta->ksi_to_psi();
      }
      accepted=accept(logprior_diff);

      for(i=0;i<num;i++){
        psi[i]->update(accepted);
      }
      if(mode==POSTMEAN){
        alpha->theta->update(accepted);
        alpha->ksi->update(accepted);
        alpha->psi->update(accepted);
        alpha->chi->update(accepted);
        beta->theta->update(accepted);
        beta->ksi->update(accepted);
        beta->psi->update(accepted);
        beta->chi->update(accepted);
      }
      num_accept+=accepted;
      run_num_accept+=accepted;
      data->update(accepted,which);  // update likelihood
    }
  }
  return(accepted);
}

void nonspatial_vecparm::add_child(spatial_parm* alpha_in,spatial_parm* beta_in){
  alpha=alpha_in;
  beta=beta_in;
}

////////////////////////////////////
// Class spatial_parm
////////////////////////////////////


spatial_parm::spatial_parm(int which_in,int init_mode, double* init,double mu_init, double logsigma_init, double logkappa_init,double mu_mean, double mu_sd, double logsigma_mean, double logsigma_sd, double logkappa_mean,double logkappa_sd, double theta_prop_sd, double theta_prop_sd_subgroup,int subgroup_size_in,double mu_prop_sd, double logsigma_prop_sd, double logkappa_prop_sd,double ratio_in,transform_fxn_ptr transform_in,double kappa_alone_propsd_in){
 
  kappa_alone_propsd=kappa_alone_propsd_in;
  num_accept_kappa2=0;
  run_num_accept_kappa2=0;

  ratio=ratio_in;
  subgroup_size=subgroup_size_in;
  int accepted=ACCEPT;
  num_accept_kappa=0;
  num_accept_theta_subgroup=0;
  num_accept_theta=0;
  num_accept_mu=0;
  num_accept_sigma=0;
  run_num_accept_kappa=0;
  run_num_accept_theta=0;
  run_num_accept_theta_subgroup=0;
  run_num_accept_mu=0;
  run_num_accept_sigma=0;
  transform=transform_in;
  which=which_in;
  num_thetapoints=locations->num_gridpoints;

  num_dim=2;

  // temporary vectors
  tmpvec=new double[num_thetapoints];
  tmpvec2=new double[num_thetapoints];

  // vectors for Langevin calculations
  oldgrad=new double[num_thetapoints];
  newgrad=new double[num_thetapoints];

  oldgrad_exceeded=0;
  newgrad_exceeded=0;

  // vectors for post mean calculations
  LCsum=new double[num_thetapoints];
  Csum_diff=new double[num_thetapoints];

  // mean of parameters - same for all locations
  mu=new hyperparm_scal(mu_init,mu_mean,mu_sd,mu_prop_sd,&normal_dens,&identity);
  // log(sd) of white noise - same for all locations
  sigma=new hyperparm_scal(logsigma_init,logsigma_mean,logsigma_sd,logsigma_prop_sd,&invgamma_dens,&log_transform);

  // put init and propsd in here and took init out of theta
  ksi=new hyperparm_vec(num_thetapoints); // ,theta_init,theta_prop_sd1);  // actual parameter vector on scale of proposals
  theta=new hyperparm_vec(locations->num_gridpoints,init,theta_prop_sd,theta_prop_sd_subgroup); //,theta_init,theta_prop_sd1);        // implicit parameter vector
  psi=new hyperparm_vec(locations->num_gridpoints);        // actual parameter vector on transformed scale: g(Kw) 
  chi=new hyperparm_vec(num_thetapoints);   // L(lambda)^-1*(f-fhat)/sigma -- only used if which==ALPHA or BETA
  kappa=new hyperparm_scal(logkappa_init,logkappa_mean,logkappa_sd,logkappa_prop_sd,&uniform_dens,&log_transform);
  kappa->evaluate_prior();                  // initialize prior densities ...
  kappa->update(accepted);                  // initialize stored values ...
   
  stat_cov=new sparse_cov_mat_stat(locations->num_gridpoints,&kappa->value);

  if(init_mode==KSI_NOREP){
    cout << "Warning - doing ksi initialization\n";
    copy(num_thetapoints,theta->value,ksi->value);
    ksi_to_theta();
    ksi_to_psi();
  }
  else{
    theta_to_ksi();  // use if not starting from scratch
    ksi_to_psi();
  }

  // no need for using random deviates for theta since I'm now starting theta at random values

  if(which==DATA_COVAR_EIGVAL){
    for(int j=0;j<locations->num_gridpoints;j++){
      if(psi->value[j]>=2.0) {
	//psi->value[j]=1.0;
	// just setting psi values doesn't change theta, so this makes no sense
	cerr << "Warning: some initial psi values for " << which << " are greater than 2.0\n";
      }
    }
  }

  cout << "Starting values for " << which << ":" << endl;
  for(int j=0;j<locations->num_gridpoints;j++){
    cout << psi->value[j] << " ";
  }
  cout << endl;

  ksi->update(accepted);
  psi->update(accepted);
  // chi not updated because it needs to be filled first
  mu->evaluate_prior();
  sigma->evaluate_prior();
  theta->evaluate_prior();
  mu->update(accepted);
  sigma->update(accepted);
  theta->update(accepted);
}

spatial_parm::~spatial_parm(){
  delete mu;
  delete sigma;
  delete theta;
  delete psi;
  delete ksi;
  delete chi;
  delete kappa;
  delete stat_cov;

  delete [] tmpvec;
  delete [] tmpvec2;

  delete [] LCsum;
  delete [] Csum_diff;

  delete [] oldgrad;
  delete [] newgrad;
}

void spatial_parm::ksi_to_theta(void){
  int i;
  cout << "Warning: ksi_to_theta function in use\n";
  for(i=0;i<num_thetapoints;i++){
    tmpvec[i]=(ksi->value[i]-mu->value)/sigma->value;
  }
  stat_cov->Linvmult(tmpvec,theta->value);
}

void spatial_parm::theta_to_ksi(void){
  stat_cov->Lmult(theta->value,tmpvec);
  for(int i=0;i<num_thetapoints;i++){
    ksi->value[i]=tmpvec[i]*sigma->value+mu->value;
  }
}

void spatial_parm::ksi_to_psi(void){
  for(int j=0;j<locations->num_gridpoints;j++){
    psi->value[j]=transform(ksi->value[j],0);
  }  
}

int spatial_parm::accept_nolik(double logprior_diff){
  float unif=ran1(idum,GENERATE,"a"); 
  if(exp(logprior_diff)>unif){ 
    return(ACCEPT);
  }
  else{
    return(REJECT);
  }
}

int spatial_parm::accept_lik(double logprior_diff){
  float unif=ran1(idum,GENERATE,"a"); 
  double loglik_diff=data->evaluate_loglik_diff(which,NOT_FIRST);
  //cout << "LL info: " << which << " " << loglik_diff << " " << logprior_diff << "\n"; cout.flush();
  if(exp(loglik_diff)*exp(logprior_diff)>unif){ 
    return(ACCEPT);
  }
  else{
    return(REJECT);
  }
}

void spatial_parm::add_child(spatial_parm* alpha_in,spatial_parm* beta_in){
  alpha=alpha_in;
  beta=beta_in;
}

void spatial_parm::add_info(void){
  if(which==ALPHA){
    stat_cov->add_info(data->get_diag_C(),data->get_lnz_C(),&(sigma->value),(double)data->num_train_years,0,0);
  }
  if(which==BETA){
    stat_cov->add_info(data->get_diag_C(),data->get_lnz_C(),&(sigma->value),(double)data->sum_t_squ,0,0);
  }
  if(which==ETA){
    stat_cov->add_info(data->get_diag_R(),data->get_lnz_R(),&(sigma->value),(double)data->num_train_years/2.0,data->etahat,data->delta);
  }
  ksi_propose(TRUE); // doesn't really propose ksi, just calculates chi
}

double spatial_parm::ksi_propose(int update_chi){
  // returns q ratio for alpha/beta proposals

  // at some point, may want to add code to save Csum and its Chol so I don't have to recalc it
  // when only mu is changed
  // mu could use saved Csum Choleski, Csum_y (actually Cparmmult*Csum_y)
  // updating chi could use Csum or LCsum

  // for fixed beta, add line if(which!=BETA)
  int i;

  // technically, for the joint sigma/kappa prop, this sd should involve the new sigma
  double sd=theta->prop_sd*.0001; 
  int flag;
  // calculate fhat 
  if(!theta->fixed){
    if(which==ALPHA){
      for(i=0;i<locations->num_gridpoints;i++){
        tmpvec[i]=(double)(data->sum_data[i])/data->num_train_years-mu->value;
      }
    }
    if(which==BETA){
      for(i=0;i<locations->num_gridpoints;i++){
        tmpvec[i]=(double)(data->sum_data_by_t[i])/data->sum_t_squ-mu->value;
      }
    }
    if(which==ETA){
      for(i=0;i<locations->num_gridpoints;i++){
        tmpvec[i]=log(data->etahat[i])-mu->value;
      }
    }
    
    stat_cov->fill_Csum(which);
    stat_cov->gen_chol_Csum();
    flag=stat_cov->solve_Csum(tmpvec,Csum_diff);
    
    if(flag){
      if(prior==STATIONARY){
        cout << "Parameters at non-invertibility: " << which << " " << sigma->value << " " << kappa->value << " " << *(data->stat_cov->noise_var) << " " << *(data->stat_cov->corr_parm) << endl;
      }
      if(prior==NON_STATIONARY){
        cout << "Parameters at non-invertibility: " << sigma->value << " " << kappa->value << " " << *(data->nonstat_cov->noise_var) << endl;
      }
      if(prior==FIXED){
        cout << "Parameters at non-invertibility: " << sigma->value << " " << kappa->value << " " << *(data->fixed_cov->noise_var) << endl;
      }
    }
    
    stat_cov->Ltmult(Csum_diff,LCsum);
    
    if(update_chi){
      for(i=0;i<num_thetapoints;i++){
        chi->value[i]=theta->value[i]-sigma->value*LCsum[i];
      }
      chi->update(ACCEPT);
    }
    else{
      stat_cov->Cparmmult(Csum_diff,ksi->value);
      chi->propose(sd);
      stat_cov->Lcheck(chi->value,chi->value_stored);  // don't update zeroed out column elements
      stat_cov->Lmult(chi->value,tmpvec);
      for(i=0;i<num_thetapoints;i++){
        ksi->value[i]+=mu->value+sigma->value*tmpvec[i];
      }
      
      // calculate proposed theta
      for(i=0;i<num_thetapoints;i++){
        theta->value[i]=chi->value[i]+sigma->value*LCsum[i];
      }    
    }
  }
  return(0);
}
  
void spatial_parm::step(int basic_mode,int langevin,int theta_prop_mode,int fastparm_fixed){
  int accepted;
  int i;
  double sd=theta->prop_sd*0.0001;
  double sigsd=0.000000001;
  double value;

  if(which!=ALPHA && which!=BETA && which!=ETA){
    // generalized S,D proposals
    // S proposal
    if(!kappa->fixed){
      accepted=REJECT;
      
      // if I ever use this code, I should probably have small amount of indep. sigma proposal noise unless I prove the result about deterministic jumps
      value=kappa->propose(0.0);
      sigma->base_value+=(1.0/ratio)*kappa->prop_sd*value;
      sigma->value=sigma->transform(sigma->base_value,0);
      logprior_diff=kappa->evaluate_prior();
      
      if(zero(logprior_diff-NEG_INFTY)){
        kappa->update(accepted);
        sigma->update(accepted);
      }
      else{
        logprior_diff+=sigma->evaluate_prior();
        stat_cov->fill(NOT_FIRST);
        stat_cov->gen_chol(NO_WARN);
        
        theta->propose(sd);
        stat_cov->Lcheck(theta->value,theta->value_stored);  // prevents updating of elements whose columns are zeroed out
        logprior_diff+=theta->evaluate_prior();
        theta_to_ksi();
        ksi_to_psi();
        
        data->datacov_change(which,NOT_FIRST);
        
        // use SIMPLE updates for eta sum proposal even when eta is POSTMEAN sampled 

        if(basic_mode==POSTMEAN){
          logprior_diff+=alpha->ksi_propose(FALSE);
          logprior_diff+=alpha->theta->evaluate_prior();
          logprior_diff+=beta->ksi_propose(FALSE);
          logprior_diff+=beta->theta->evaluate_prior();
          alpha->ksi_to_psi();
          beta->ksi_to_psi();
        }
        
        accepted=accept_lik(logprior_diff);
        
        // use SIMPLE updates for eta sum proposal even when eta is POSTMEAN sampled 
        
        if(basic_mode==POSTMEAN){
          alpha->theta->update(accepted);
          alpha->ksi->update(accepted);
          alpha->psi->update(accepted);
          alpha->chi->update(accepted);
          beta->theta->update(accepted);
          beta->ksi->update(accepted);
          beta->psi->update(accepted);
          beta->chi->update(accepted);
        }

        ksi->update(accepted);
        psi->update(accepted);
        data->update(accepted,which);
        kappa->update(accepted);
        sigma->update(accepted);
        theta->update(accepted);
        stat_cov->update(accepted);
        
        num_accept_kappa+=accepted;
        run_num_accept_kappa+=accepted;
      }
    }
    // D proposal 
    if(0==1){
      accepted=REJECT;
      value=sigma->propose(kappa->prop_sd);
      kappa->base_value-=ratio*kappa->prop_sd*value;
      kappa->value=kappa->transform(kappa->base_value,0);
      logprior_diff=kappa->evaluate_prior();
      if(logprior_diff<(NEG_INFTY+.0001) && logprior_diff>(NEG_INFTY-.0001)){
        kappa->update(accepted);
        sigma->update(accepted);
      }
      else{
      logprior_diff+=sigma->evaluate_prior();
      
      stat_cov->fill(NOT_FIRST);
      stat_cov->gen_chol(NO_WARN);

      theta->propose(sd);
      stat_cov->Lcheck(theta->value,theta->value_stored);  // prevents updating of elements whose columns are zeroed out
      logprior_diff+=theta->evaluate_prior();
      theta_to_ksi();
      ksi_to_psi();
      
      data->datacov_change(which,NOT_FIRST);

      if(basic_mode==POSTMEAN){
        logprior_diff+=alpha->ksi_propose(FALSE);
        logprior_diff+=alpha->theta->evaluate_prior();
        logprior_diff+=beta->ksi_propose(FALSE);
        logprior_diff+=beta->theta->evaluate_prior();
        alpha->ksi_to_psi();
        beta->ksi_to_psi();
      }
      
      accepted=accept_lik(logprior_diff);
      
      if(basic_mode==POSTMEAN){
        alpha->theta->update(accepted);
        alpha->ksi->update(accepted);
        alpha->psi->update(accepted);
        alpha->chi->update(accepted);
        beta->theta->update(accepted);
        beta->ksi->update(accepted);
        beta->psi->update(accepted);
        beta->chi->update(accepted);
      }

      ksi->update(accepted);
      psi->update(accepted);
      data->update(accepted,which);
      kappa->update(accepted);
      sigma->update(accepted);
      theta->update(accepted);
      stat_cov->update(accepted);
      
      num_accept_kappa+=accepted;
      run_num_accept_kappa+=accepted;
    }
    }
    // joint prop
    if(0==1 && which==ETA && !kappa->fixed && !sigma->fixed){
      accepted=REJECT;
      value=kappa->propose(0.0);
      logprior_diff=kappa->evaluate_prior();
      
      if(logprior_diff<(NEG_INFTY+.0001) && logprior_diff>(NEG_INFTY-.0001)){
        kappa->update(accepted);
      }
      else{	
        sigma->propose(.8*sigma->prop_sd,0.6*value);
        logprior_diff+=sigma->evaluate_prior();
        
        stat_cov->fill(NOT_FIRST);
        stat_cov->gen_chol(NO_WARN);
        theta->propose(sd);
        stat_cov->Lcheck(theta->value,theta->value_stored);  // prevents updating of elements whose columns are zeroed out
        logprior_diff+=theta->evaluate_prior();
        theta_to_ksi();
        ksi_to_psi();
        
        data->datacov_change(which,NOT_FIRST);
        
	if(basic_mode==POSTMEAN){
	  logprior_diff+=alpha->ksi_propose(FALSE);
	  logprior_diff+=alpha->theta->evaluate_prior();
	  logprior_diff+=beta->ksi_propose(FALSE);
	  logprior_diff+=beta->theta->evaluate_prior();
	  alpha->ksi_to_psi();
	  beta->ksi_to_psi();
	}
	
        accepted=accept_lik(logprior_diff);
        
        if(basic_mode==POSTMEAN){
          alpha->theta->update(accepted);
          alpha->ksi->update(accepted);
          alpha->psi->update(accepted);
          alpha->chi->update(accepted);
          beta->theta->update(accepted);
          beta->ksi->update(accepted);
          beta->psi->update(accepted);
          beta->chi->update(accepted);
        }
        
        ksi->update(accepted);
        psi->update(accepted);
        data->update(accepted,which);
        kappa->update(accepted);
        sigma->update(accepted);
        theta->update(accepted);
        stat_cov->update(accepted);
        
        num_accept_sigma+=accepted;
        run_num_accept_sigma+=accepted;
        
        num_accept_kappa+=accepted;
        run_num_accept_kappa+=accepted;
      }
    }
    // individual proposal
    if(!kappa->fixed){
      if(0==1){
        accepted=REJECT;
        logprior_diff=0.0;
        
        kappa->propose(0.0);	
        logprior_diff+=kappa->evaluate_prior();
        
        if(logprior_diff<(NEG_INFTY+.0001) && logprior_diff>(NEG_INFTY-.0001)){
          kappa->update(accepted);
        }
        else{
          stat_cov->fill(NOT_FIRST);
          stat_cov->gen_chol(NO_WARN);
          theta->propose(sd);
          stat_cov->Lcheck(theta->value,theta->value_stored);  // prevents updating of elements whose columns are zeroed out
          logprior_diff+=theta->evaluate_prior();
          theta_to_ksi();
          ksi_to_psi();
          data->datacov_change(which,NOT_FIRST);
          // this needs to be here and not in eval_loglik because ksi_propose makes use of data covariance
          
          if(basic_mode==POSTMEAN){
            logprior_diff+=alpha->ksi_propose(FALSE);
            logprior_diff+=alpha->theta->evaluate_prior();
            logprior_diff+=beta->ksi_propose(FALSE);
            logprior_diff+=beta->theta->evaluate_prior();
            alpha->ksi_to_psi();
            beta->ksi_to_psi();
          }
          
          accepted=accept_lik(logprior_diff);
          /* this is for full kernel approach when eigval is normal rather than uniform
             if(which==DATA_COVAR_EIGVAL){
             for(int i=0;i<locations->num_gridpoints;i++){
             if(psi->value[i]>2.0){
             accepted=REJECT;
             }
             }
             }
          */
          if(basic_mode==POSTMEAN){
            alpha->theta->update(accepted);
            alpha->ksi->update(accepted);
            alpha->psi->update(accepted);
            alpha->chi->update(accepted);
            beta->theta->update(accepted);
            beta->ksi->update(accepted);
            beta->psi->update(accepted);
            beta->chi->update(accepted);
          }
          
          kappa->update(accepted);
          stat_cov->update(accepted);
          theta->update(accepted);
          ksi->update(accepted);
          psi->update(accepted);
          data->update(accepted,which);
          
          num_accept_kappa+=accepted;
          run_num_accept_kappa+=accepted;
        }
      }
    }
    // separate mu, sigma and theta proposals
    
    
    if(!mu->fixed){
      mu->propose(0.0);
      logprior_diff=mu->evaluate_prior();
      theta->propose(sd);
      stat_cov->Lcheck(theta->value,theta->value_stored);
      logprior_diff+=theta->evaluate_prior();
      theta_to_ksi();
      ksi_to_psi();
      data->datacov_change(which,NOT_FIRST);
      
      if(basic_mode==POSTMEAN){
        logprior_diff+=alpha->ksi_propose(FALSE);
        logprior_diff+=alpha->theta->evaluate_prior();
        logprior_diff+=beta->ksi_propose(FALSE);
        logprior_diff+=beta->theta->evaluate_prior();
        alpha->ksi_to_psi();
        beta->ksi_to_psi();
      }

      accepted=accept_lik(logprior_diff);
      /* for full kernel approach
         if(which==DATA_COVAR_EIGVAL){
         for(int i=0;i<locations->num_gridpoints;i++){
         if(psi->value[i]>2.0){
         accepted=REJECT;
         }
         }
         }
      */
      
      if(basic_mode==POSTMEAN){
        alpha->theta->update(accepted);
        alpha->ksi->update(accepted);
        alpha->psi->update(accepted);
        alpha->chi->update(accepted);
        beta->theta->update(accepted);
        beta->ksi->update(accepted);
        beta->psi->update(accepted);
        beta->chi->update(accepted);
      }
      
      mu->update(accepted);
      theta->update(accepted);
      ksi->update(accepted);
      psi->update(accepted);
      data->update(accepted,which);
      
      num_accept_mu+=accepted;
      run_num_accept_mu+=accepted;
    }
    if(!sigma->fixed){
      // POSTMEAN sampling for eta, i.e. simple joint for sigma,theta
      if(1==1){
        sigma->propose(0.0);
        logprior_diff=sigma->evaluate_prior();

        theta->propose(sd);
        stat_cov->Lcheck(theta->value,theta->value_stored);  // prevents updating of elements whose columns are zeroed out
        theta_to_ksi();
        ksi_to_psi();
        logprior_diff+=theta->evaluate_prior();
        
        data->datacov_change(which,NOT_FIRST);

        if(basic_mode==POSTMEAN){
          logprior_diff+=alpha->ksi_propose(FALSE);
          logprior_diff+=alpha->theta->evaluate_prior();
          logprior_diff+=beta->ksi_propose(FALSE);
          logprior_diff+=beta->theta->evaluate_prior();
          alpha->ksi_to_psi();
          beta->ksi_to_psi();
        }

        accepted=accept_lik(logprior_diff);
        /* for full kernel approach
           if(which==DATA_COVAR_EIGVAL){
           for(int i=0;i<locations->num_gridpoints;i++){
           if(psi->value[i]>2.0){
           accepted=REJECT;
           }
           }
           }
        */

        if(basic_mode==POSTMEAN){
          alpha->theta->update(accepted);
          alpha->ksi->update(accepted);
          alpha->psi->update(accepted);
          alpha->chi->update(accepted);
          beta->theta->update(accepted);
          beta->ksi->update(accepted);
          beta->psi->update(accepted);
          beta->chi->update(accepted);
        }

        ksi->update(accepted);
        psi->update(accepted);
        data->update(accepted,which);
        sigma->update(accepted);
        theta->update(accepted);
        num_accept_sigma+=accepted;
        run_num_accept_sigma+=accepted;
      }
      // hybridized
      if(0==1){
        sigma->propose(0.0);
        logprior_diff=sigma->evaluate_prior();
        for(i=0;i<num_thetapoints;i++){
          theta->value[i]*=sigma->value_stored/sigma->value;
        }
        logprior_diff+=theta->evaluate_prior();
        logprior_diff+=num_thetapoints*(log(sigma->value_stored)-log(sigma->value));
        accepted=accept_nolik(logprior_diff);

        // no need for alpha/beta chi update since don't change eta psi
        sigma->update(accepted);
        theta->update(accepted);
        num_accept_sigma+=accepted;
        run_num_accept_sigma+=accepted;
      }
    }
    if(!theta->fixed){
      if(!langevin){
        theta->propose(theta->prop_sd/sigma->value);
        stat_cov->Lcheck(theta->value,theta->value_stored);  // prevents updating of elements whose columns are zeroed out
        logprior_diff=theta->evaluate_prior();   // Mahl_dist and theta are same for hyperparms
        theta_to_ksi();
        ksi_to_psi();
        data->datacov_change(which,NOT_FIRST);
        
        if(basic_mode==POSTMEAN){
          logprior_diff+=alpha->ksi_propose(FALSE);
          logprior_diff+=alpha->theta->evaluate_prior();
          logprior_diff+=beta->ksi_propose(FALSE);
          logprior_diff+=beta->theta->evaluate_prior();
          alpha->ksi_to_psi();
          beta->ksi_to_psi();
        }
        
        accepted=accept_lik(logprior_diff);
        /* for full kernel approach
           if(which==DATA_COVAR_EIGVAL){
           for(int i=0;i<locations->num_gridpoints;i++){
           if(psi->value[i]>2.0){
           accepted=REJECT;
           }
           }
           }
        */
        
        if(basic_mode==POSTMEAN){
          alpha->theta->update(accepted);
          alpha->ksi->update(accepted);
          alpha->psi->update(accepted);
          alpha->chi->update(accepted);
          beta->theta->update(accepted);
          beta->ksi->update(accepted);
          beta->psi->update(accepted);
          beta->chi->update(accepted);
        }
        
        theta->update(accepted);
        ksi->update(accepted);
        psi->update(accepted);
        
        data->update(accepted,which);
        
        num_accept_theta+=accepted;
        run_num_accept_theta+=accepted;
        
        if(!theta->subgroup_fixed){  // sample first few thetas separately
          theta->propose(theta->prop_sd_subgroup/sigma->value,subgroup_size);
          stat_cov->Lcheck(theta->value,theta->value_stored);  // prevents updating of elements whose columns are zeroed out
          logprior_diff=theta->evaluate_prior();   // Mahl_dist and theta are same for hyperparms
          theta_to_ksi();
          ksi_to_psi();
          data->datacov_change(which,NOT_FIRST);
          
          if(basic_mode==POSTMEAN){
            logprior_diff+=alpha->ksi_propose(FALSE);
            logprior_diff+=alpha->theta->evaluate_prior();
            logprior_diff+=beta->ksi_propose(FALSE);
            logprior_diff+=beta->theta->evaluate_prior();
            alpha->ksi_to_psi();
            beta->ksi_to_psi();
          }
          
          accepted=accept_lik(logprior_diff);
          /* for full kernel approach 
             if(which==DATA_COVAR_EIGVAL){
             for(int i=0;i<locations->num_gridpoints;i++){
             if(psi->value[i]>2.0){
             accepted=REJECT;
             }
             }
             }
          */
          if(basic_mode==POSTMEAN){
            alpha->theta->update(accepted);
            alpha->ksi->update(accepted);
            alpha->psi->update(accepted);
            alpha->chi->update(accepted);
            beta->theta->update(accepted);
            beta->ksi->update(accepted);
            beta->psi->update(accepted);
            beta->chi->update(accepted);
          }
          
          
          theta->update(accepted);
          ksi->update(accepted);
          psi->update(accepted);
          
          data->update(accepted,which);
          
          num_accept_theta_subgroup+=accepted;
          run_num_accept_theta_subgroup+=accepted;
        }
      }
      else{
        cerr << "Warning: langevin sampling for parameters other than alpha and beta not implemented\n";
      }
    }
  }
  else{  // ALPHA or BETA or ETA (now that I have approx PM for eta) sampling
    // generalized S,D proposals
    // S proposal
    if(!kappa->fixed){
      accepted=REJECT;
      value=kappa->propose(0.0);
      sigma->propose(sigsd);
      sigma->base_value+=(1.0/ratio)*kappa->prop_sd*value;
      sigma->value=sigma->transform(sigma->base_value,0);
      logprior_diff=kappa->evaluate_prior();
      if(zero(logprior_diff-NEG_INFTY)){
        kappa->update(accepted);
        sigma->update(accepted);
      }
      else{
        logprior_diff+=sigma->evaluate_prior();
        
        stat_cov->fill(NOT_FIRST);
        stat_cov->gen_chol(NO_WARN);

        // technically, the proposal for the process should have the new sigma involved in the proposal variance, so that the Hastings ratio is done correctly; with the small prop variance, it doesn't really make a practical difference
        if(basic_mode==POSTMEAN){
          logprior_diff+=ksi_propose(FALSE);
        }
        else{
          theta->propose(sd);
          stat_cov->Lcheck(theta->value,theta->value_stored);  // prevents updating of elements whose columns are zeroed out
          theta_to_ksi();
        }
        logprior_diff+=theta->evaluate_prior();
        ksi_to_psi();
        
        if(which==ETA){
          data->datacov_change(which,NOT_FIRST);
        }

        /* not clear this helps much, but otherwise, need to do chi update if accepted anyway */
        /*
        if(which==ETA && basic_mode==POSTMEAN){
          logprior_diff+=alpha->ksi_propose(FALSE);
          logprior_diff+=alpha->theta->evaluate_prior();
          logprior_diff+=beta->ksi_propose(FALSE);
          logprior_diff+=beta->theta->evaluate_prior();
          alpha->ksi_to_psi();
          beta->ksi_to_psi();
        }
        */

        accepted=accept_lik(logprior_diff);
        /* for full kernel approach
           if(which==DATA_COVAR_EIGVAL){
           for(int i=0;i<locations->num_gridpoints;i++){
           if(psi->value[i]>2.0){
           accepted=REJECT;
           }
           }
           }
        */

        /*
        if(which==ETA && basic_mode==POSTMEAN){
          alpha->theta->update(accepted);
          alpha->ksi->update(accepted);
          alpha->psi->update(accepted);
          alpha->chi->update(accepted);
          beta->theta->update(accepted);
          beta->ksi->update(accepted);
          beta->psi->update(accepted);
          beta->chi->update(accepted);
        }
        */

        if(basic_mode==POSTMEAN){
          chi->update(accepted);
        }
        
        ksi->update(accepted);
        psi->update(accepted);
        data->update(accepted,which);
        kappa->update(accepted);
        sigma->update(accepted);
        theta->update(accepted);
        stat_cov->update(accepted);
        
        num_accept_kappa+=accepted;
        run_num_accept_kappa+=accepted;
      }
    }


    // propose kappa on its own when running chain from widely-dispersed starting points; this helps with mixing
    if(!kappa->fixed && !zero(kappa_alone_propsd)){
      accepted=REJECT;
      value=kappa->propose(kappa_alone_propsd);
      logprior_diff=kappa->evaluate_prior();
      if(zero(logprior_diff-NEG_INFTY)){
        kappa->update(accepted);
      }
      else{
        stat_cov->fill(NOT_FIRST);
        stat_cov->gen_chol(NO_WARN);
        
        if(basic_mode==POSTMEAN){
          logprior_diff+=ksi_propose(FALSE);
        }
        else{
          theta->propose(sd);
          stat_cov->Lcheck(theta->value,theta->value_stored);  // prevents updating of elements whose columns are zeroed out
          theta_to_ksi();
        }
        logprior_diff+=theta->evaluate_prior();
        ksi_to_psi();
        
        if(which==ETA){
          data->datacov_change(which,NOT_FIRST);
        }

        accepted=accept_lik(logprior_diff);

        if(basic_mode==POSTMEAN){
          chi->update(accepted);
        }
        
        ksi->update(accepted);
        psi->update(accepted);
        data->update(accepted,which);
        kappa->update(accepted);
        theta->update(accepted);
        stat_cov->update(accepted);
        
        num_accept_kappa2+=accepted;
        run_num_accept_kappa2+=accepted;
      }
    }

      // D proposal; not used since doing (kappa,sigma),(sigma) proposals
    if(0==1){
    accepted=REJECT;
    value=sigma->propose(kappa->prop_sd);
    kappa->base_value-=ratio*kappa->prop_sd*value;
    kappa->value=kappa->transform(kappa->base_value,0);
    logprior_diff=kappa->evaluate_prior();
    if(logprior_diff<(NEG_INFTY+.0001) && logprior_diff>(NEG_INFTY-.0001)){
      kappa->update(accepted);
      sigma->update(accepted);
    }
    else{
      logprior_diff+=sigma->evaluate_prior();
      
      stat_cov->fill(NOT_FIRST);
      stat_cov->gen_chol(NO_WARN);

      if(basic_mode==POSTMEAN){
        logprior_diff+=ksi_propose(FALSE);
      }
      else{
        theta->propose(sd);
        stat_cov->Lcheck(theta->value,theta->value_stored);  // prevents updating of elements whose columns are zeroed out
        theta_to_ksi();
      }
      logprior_diff+=theta->evaluate_prior();
      ksi_to_psi();
      
      if(which==ETA){
        data->datacov_change(which,NOT_FIRST);
      }

      accepted=accept_lik(logprior_diff);

      if(basic_mode==POSTMEAN){
        chi->update(accepted);
      }
      
      ksi->update(accepted);
      psi->update(accepted);
      data->update(accepted,which);
      kappa->update(accepted);
      sigma->update(accepted);
      theta->update(accepted);
      stat_cov->update(accepted);
      
      num_accept_kappa+=accepted;
      run_num_accept_kappa+=accepted;
    }
    }

    /* not used since doing (kappa,sigma), (sigma) proposals */
    if(!kappa->fixed){
      // individual kappa sampling
      if(0==1){
        accepted=REJECT;
        if(kappa->value>kappa_highval){
          kappa->propose(kappa->prop_sd*highkappa_scale);
        }
        else{
          kappa->propose(0.0);
        }
        logprior_diff=kappa->evaluate_prior();
        if(logprior_diff<(NEG_INFTY+.0001) && logprior_diff>(NEG_INFTY-.0001)){
          kappa->update(accepted);
        }
        else{
          if(kappa->value>kappa_highval && kappa->value_stored<=kappa_highval){
            logprior_diff+=(-log(highkappa_scale))-(pow(kappa->base_value-kappa->base_value_stored,2.0)/(2*kappa->prop_sd*kappa->prop_sd))*(1/(highkappa_scale*highkappa_scale)-1);
          }
          if(kappa->value<=kappa_highval && kappa->value_stored>kappa_highval){
            logprior_diff+=(log(highkappa_scale))-(pow(kappa->base_value-kappa->base_value_stored,2.0)/(2.0*kappa->prop_sd*kappa->prop_sd))*(1.0-1.0/(highkappa_scale*highkappa_scale));
          }
          stat_cov->fill(NOT_FIRST);
          stat_cov->gen_chol(NO_WARN);
          
          if(basic_mode==POSTMEAN){
            logprior_diff+=ksi_propose(FALSE);
          }
          else{  // SIMPLE
            theta->propose(sd);
            stat_cov->Lcheck(theta->value,theta->value_stored);  // prevents updating of elements whose columns are zeroed out
            theta_to_ksi();
          }
          
          logprior_diff+=theta->evaluate_prior();
          
          ksi_to_psi();
          
          if(which==ETA){
            data->datacov_change(which,NOT_FIRST);
          }
          
          accepted=accept_lik(logprior_diff);
          
          if(basic_mode==POSTMEAN){
            chi->update(accepted);
          }
	  
          kappa->update(accepted);
          stat_cov->update(accepted);
          theta->update(accepted);
          ksi->update(accepted);
          psi->update(accepted);
          data->update(accepted,which);
          
          num_accept_kappa+=accepted;
          run_num_accept_kappa+=accepted;
        }
      }
    }
    if(!mu->fixed && !fastparm_fixed){
      mu->propose(0.0);
      logprior_diff=mu->evaluate_prior();
      
      if(basic_mode==POSTMEAN){
        logprior_diff+=ksi_propose(FALSE);
      }
      else{  // SIMPLE
        theta->propose(sd);
        stat_cov->Lcheck(theta->value,theta->value_stored);
        theta_to_ksi();
      }
      logprior_diff+=theta->evaluate_prior();
      ksi_to_psi();
      
      if(which==ETA){
        data->datacov_change(which,NOT_FIRST);
      }

      /*
      if(which==ETA && basic_mode==POSTMEAN){
        logprior_diff+=alpha->ksi_propose(FALSE);
        logprior_diff+=alpha->theta->evaluate_prior();
        logprior_diff+=beta->ksi_propose(FALSE);
        logprior_diff+=beta->theta->evaluate_prior();
        alpha->ksi_to_psi();
        beta->ksi_to_psi();
      }
      */

      accepted=accept_lik(logprior_diff);
      /* for full kernel approach
             if(which==DATA_COVAR_EIGVAL){
             for(int i=0;i<locations->num_gridpoints;i++){
             if(psi->value[i]>2.0){
             accepted=REJECT;
             }
             }
             }
          */
        
      /*
      if(which==ETA && basic_mode==POSTMEAN){
        alpha->theta->update(accepted);
        alpha->ksi->update(accepted);
        alpha->psi->update(accepted);
        alpha->chi->update(accepted);
        beta->theta->update(accepted);
        beta->ksi->update(accepted);
        beta->psi->update(accepted);
        beta->chi->update(accepted);
      }
      */
      
      if(basic_mode==POSTMEAN){
        chi->update(accepted);
      }
      mu->update(accepted);
      theta->update(accepted);
      ksi->update(accepted);
      psi->update(accepted);
      data->update(accepted,which);
      num_accept_mu+=accepted;
      run_num_accept_mu+=accepted;
    }
        
    if(!sigma->fixed){  
      // hybridized
      if(1==1){
     if(which==ALPHA || which==BETA || which==ETA){
        sigma->propose(0.0);
        logprior_diff=sigma->evaluate_prior();
        for(i=0;i<num_thetapoints;i++){
          theta->value[i]*=sigma->value_stored/sigma->value;
        }
        logprior_diff+=theta->evaluate_prior();
        logprior_diff+=num_thetapoints*(log(sigma->value_stored)-log(sigma->value));
        accepted=accept_nolik(logprior_diff);
        
        if(accepted && basic_mode==POSTMEAN){
          ksi_propose(TRUE);
        }
        
        
        sigma->update(accepted);
        theta->update(accepted);
        
        num_accept_sigma+=accepted;
        run_num_accept_sigma+=accepted;
      }
      }
      // simple (joint)
      if(0==1){
        sigma->propose(0.0);
        logprior_diff=sigma->evaluate_prior();
        
        theta->propose(sd);
        stat_cov->Lcheck(theta->value,theta->value_stored);  // prevents updating of elements whose columns are zeroed out
        theta_to_ksi();
        ksi_to_psi();
        
        logprior_diff+=theta->evaluate_prior();
        
        if(which==ETA){
          data->datacov_change(which,NOT_FIRST);
        }
        
        accepted=accept_lik(logprior_diff);
        
        if(accepted && basic_mode==POSTMEAN){
          ksi_propose(TRUE);
        }
        
        ksi->update(accepted);
        psi->update(accepted);
        data->update(accepted,which);
        sigma->update(accepted);
        theta->update(accepted);
        
        num_accept_sigma+=accepted;
        run_num_accept_sigma+=accepted;
      }
      // postmean	
      if(0==1){ // which==ETA
        sigma->propose(0.0);
        logprior_diff=sigma->evaluate_prior();
        
        logprior_diff+=ksi_propose(FALSE);
        logprior_diff+=theta->evaluate_prior();
        ksi_to_psi();
        
        if(which==ETA){
          data->datacov_change(which,NOT_FIRST);
        }
        
        /*
        if(which==ETA && basic_mode==POSTMEAN){
          logprior_diff+=alpha->ksi_propose(FALSE);
          logprior_diff+=alpha->theta->evaluate_prior();
          logprior_diff+=beta->ksi_propose(FALSE);
          logprior_diff+=beta->theta->evaluate_prior();
          alpha->ksi_to_psi();
          beta->ksi_to_psi();
        }
        */
        accepted=accept_lik(logprior_diff);
        /* for full kernel approach
           if(which==DATA_COVAR_EIGVAL){
           for(int i=0;i<locations->num_gridpoints;i++){
           if(psi->value[i]>2.0){
           accepted=REJECT;
           }
           }
           }
        */
        /*
        if(which==ETA && basic_mode==POSTMEAN){
          alpha->theta->update(accepted);
          alpha->ksi->update(accepted);
          alpha->psi->update(accepted);
          alpha->chi->update(accepted);
          beta->theta->update(accepted);
          beta->ksi->update(accepted);
          beta->psi->update(accepted);
          beta->chi->update(accepted);
        }
        */
        ksi->update(accepted);
        psi->update(accepted);
        chi->update(accepted);
        data->update(accepted,which);
        sigma->update(accepted);
        theta->update(accepted);
        
        num_accept_sigma+=accepted;
        run_num_accept_sigma+=accepted;
      }
    }    
    // joint kappa-sigma proposal
    if(0==1 && !(kappa->fixed || sigma->fixed)){
      accepted=REJECT;
      if(kappa->value>kappa_highval){
        value=kappa->propose(kappa->prop_sd*highkappa_scale);
      }
      else{
        value=kappa->propose(0.0);
      }
      logprior_diff=kappa->evaluate_prior();
      if(logprior_diff<(NEG_INFTY+.0001) && logprior_diff>(NEG_INFTY-.0001)){
        kappa->update(accepted);
      }
      else{
        // this mess increases prop_sd if kappa is large
        if(kappa->value>kappa_highval && kappa->value_stored<=kappa_highval){
          logprior_diff+=(-log(highkappa_scale))-(pow(kappa->base_value-kappa->base_value_stored,2.0)/(2*kappa->prop_sd*kappa->prop_sd))*(1/(highkappa_scale*highkappa_scale)-1);
        }
        if(kappa->value<=kappa_highval && kappa->value_stored>kappa_highval){
          logprior_diff+=(log(highkappa_scale))-(pow(kappa->base_value-kappa->base_value_stored,2.0)/(2.0*kappa->prop_sd*kappa->prop_sd))*(1.0-1.0/(highkappa_scale*highkappa_scale));
        }
        stat_cov->fill(NOT_FIRST);
        stat_cov->gen_chol(NO_WARN);
        
        if(which==BETA){
          sigma->propose(.714*sigma->prop_sd,0.7*value);
        }
        if(which==ALPHA){
          sigma->propose(.714*sigma->prop_sd,0.7*value);
          //sigma->propose(.436*sigma->prop_sd,0.9*value);
        }
        
        logprior_diff+=sigma->evaluate_prior();
        if(basic_mode==POSTMEAN){
          logprior_diff+=ksi_propose(FALSE);
        }
        else{
          theta->propose(sd);
          stat_cov->Lcheck(theta->value,theta->value_stored);  // prevents updating of elements whose columns are zeroed out
          theta_to_ksi();
        }
        
        logprior_diff+=theta->evaluate_prior();
        ksi_to_psi();
        
        if(which==ETA){
          data->datacov_change(which,NOT_FIRST);
        }
        
        accepted=accept_lik(logprior_diff);
        
        if(basic_mode==POSTMEAN){
          chi->update(accepted);
        }
        ksi->update(accepted);
        psi->update(accepted);
        chi->update(accepted);
        data->update(accepted,which);
        kappa->update(accepted);
        sigma->update(accepted);
        theta->update(accepted);
        stat_cov->update(accepted);
        
        num_accept_sigma+=accepted;
        run_num_accept_sigma+=accepted;
        
        num_accept_kappa+=accepted;
        run_num_accept_kappa+=accepted;
      }
    }
    if(!theta->fixed){
      if(!langevin){  // non-Langevin
        
        theta->propose(theta->prop_sd/sigma->value);
        //theta->propose(0.0);
        
        stat_cov->Lcheck(theta->value,theta->value_stored);
        logprior_diff=theta->evaluate_prior();
        
        theta_to_ksi();
        ksi_to_psi();
        
        if(which==ETA){
          data->datacov_change(which,NOT_FIRST);
        }

        /*
        if(which==ETA && basic_mode==POSTMEAN){
          logprior_diff+=alpha->ksi_propose(FALSE);
          logprior_diff+=alpha->theta->evaluate_prior();
          logprior_diff+=beta->ksi_propose(FALSE);
          logprior_diff+=beta->theta->evaluate_prior();
          alpha->ksi_to_psi();
          beta->ksi_to_psi();
        }
        */
        
        accepted=accept_lik(logprior_diff);
        /* for full kernel approach
           if(which==DATA_COVAR_EIGVAL){
           for(int i=0;i<locations->num_gridpoints;i++){
           if(psi->value[i]>2.0){
           accepted=REJECT;
           }
           }
           }
        */
        
        /*
        if(which==ETA && basic_mode==POSTMEAN){
          alpha->theta->update(accepted);
          alpha->ksi->update(accepted);
          alpha->psi->update(accepted);
          alpha->chi->update(accepted);
          beta->theta->update(accepted);
          beta->ksi->update(accepted);
          beta->psi->update(accepted);
          beta->chi->update(accepted);
        }
        */
        
        if(basic_mode==POSTMEAN && accepted){
          for(i=0;i<num_thetapoints;i++){
            chi->value[i]+=theta->value[i]-theta->value_stored[i];  
          }
          chi->update(accepted);
        }
        
        theta->update(accepted);
        ksi->update(accepted);
        psi->update(accepted);
        data->update(accepted,which);  
        num_accept_theta+=accepted;
        run_num_accept_theta+=accepted;
        
        if(!theta->subgroup_fixed){
          theta->propose(theta->prop_sd_subgroup/sigma->value,subgroup_size);
          stat_cov->Lcheck(theta->value,theta->value_stored);
          logprior_diff=theta->evaluate_prior();
          
          theta_to_ksi();
          ksi_to_psi();
          
          if(which==ETA){
            data->datacov_change(which,NOT_FIRST);
          }
          
          accepted=accept_lik(logprior_diff);
          
          if(basic_mode==POSTMEAN && accepted){
            for(i=0;i<num_thetapoints;i++){
              chi->value[i]+=theta->value[i]-theta->value_stored[i];  
            }
            chi->update(accepted);
          }
          
          theta->update(accepted);
          ksi->update(accepted);
          psi->update(accepted);
          data->update(accepted,which);  // update likelihood and data covariance matrix  -- but only for function
          num_accept_theta_subgroup+=accepted;
          run_num_accept_theta_subgroup+=accepted;
        }
      }
      
      else{    // Langevin
        double H=50.0;
        if(which==ETA){
          cout << "Warning: Langevin sampling for eta not coded\n";
        }
        
        // gradient at current values for use in proposal mean
        if(which==ALPHA){
          for(i=0;i<num_thetapoints;i++){
            tmpvec[i]=data->sum_data[i]-data->num_train_years*psi->value[i];
          }
        }
        else{  // BETA
          for(i=0;i<num_thetapoints;i++){
            tmpvec[i]=data->sum_data_by_t[i]-data->sum_t_squ*psi->value[i];
          }
        }
        if(prior==STATIONARY){
          data->stat_cov->solve(tmpvec,tmpvec2);
        }
        if(prior==NON_STATIONARY){
          data->nonstat_cov->solve(tmpvec,tmpvec2);
        }
        if(prior==FIXED){
          data->fixed_cov->solve(tmpvec,tmpvec2);
        }
        stat_cov->Ltmult(tmpvec2,oldgrad);
        int flag=0;
        for(i=0;i<num_thetapoints;i++){
          oldgrad[i]=sigma->value*oldgrad[i]-theta->value[i];
          if(oldgrad[i]>H){
            flag=1;
            oldgrad[i]=H;
          }
          if(oldgrad[i]<(-H)){
            oldgrad[i]=-H;
            flag=1;
          }
        }
        if(flag){
	  oldgrad_exceeded++;
        }
        theta->propose(theta->prop_sd/sigma->value);  // random noise
        stat_cov->Lcheck(theta->value,theta->value_stored);
        
        for(i=0;i<num_thetapoints;i++){  // here's the proposal mean
          theta->value[i]+=(theta->prop_sd*theta->prop_sd)*oldgrad[i]/(2.0*sigma->value);
        }
        if(basic_mode==POSTMEAN){
          for(i=0;i<num_thetapoints;i++){  // here's the proposal mean	  
            chi->value[i]+=theta->value[i]-theta->value_stored[i];
          }
        }
        
        
        theta_to_ksi();
        ksi_to_psi();
        
        // gradient at proposed value for use in Hastings ratio
        if(which==ALPHA){
          for(i=0;i<num_thetapoints;i++){
            tmpvec[i]=data->sum_data[i]-data->num_train_years*psi->value[i];
          }
        }
        else{
          for(i=0;i<num_thetapoints;i++){
            tmpvec[i]=data->sum_data_by_t[i]-data->sum_t_squ*psi->value[i];
          }
        }
        data->stat_cov->solve(tmpvec,tmpvec2);
        stat_cov->Ltmult(tmpvec2,newgrad);
        flag=0;
        for(i=0;i<num_thetapoints;i++){
          newgrad[i]=sigma->value*newgrad[i]-theta->value[i];
          if(newgrad[i]>H){
            newgrad[i]=H;
            flag=1;
          }
          if(newgrad[i]<(-H)){
            newgrad[i]=-H;
            flag=1;
          }
        }
        if(flag){
          newgrad_exceeded++;
        }
        logprior_diff=theta->evaluate_prior();
        
        for(i=0;i<num_thetapoints;i++){
          tmpvec[i]=sigma->value*(theta->value_stored[i]-theta->value[i])-(theta->prop_sd*theta->prop_sd/2.0)*newgrad[i];
          tmpvec2[i]=sigma->value*(theta->value[i]-theta->value_stored[i])-(theta->prop_sd*theta->prop_sd/2.0)*oldgrad[i];
          logprior_diff+=-0.5*(tmpvec[i]*tmpvec[i]-tmpvec2[i]*tmpvec2[i])/(theta->prop_sd*theta->prop_sd);
        }
        
        // 1/18/03: why is this here, I can't do Langevin for eta anyway?
        if(which==ETA){
          data->datacov_change(which,NOT_FIRST);
        }
        
        accepted=accept_lik(logprior_diff);
        
        theta->update(accepted);
        ksi->update(accepted);
        psi->update(accepted);
        
        if(basic_mode==POSTMEAN){
          chi->update(accepted);
        }
        
        data->update(accepted,which);  // update likelihood and data covariance matrix  -- but only for function
        num_accept_theta+=accepted;
        run_num_accept_theta+=accepted;
        
        if(!theta->subgroup_fixed){
          double Hsub=50.0;
          
          // gradient at current values for use in proposal mean
          if(which==ALPHA){
            for(i=0;i<subgroup_size;i++){
              tmpvec[i]=data->sum_data[i]-data->num_train_years*psi->value[i];
            }
          }
          else{  // BETA
            for(i=0;i<subgroup_size;i++){
              tmpvec[i]=data->sum_data_by_t[i]-data->sum_t_squ*psi->value[i];
            }
          }
          if(prior==STATIONARY){
            data->stat_cov->solve(tmpvec,tmpvec2);
          }
          if(prior==STATIONARY){
            data->nonstat_cov->solve(tmpvec,tmpvec2);
          }
          if(prior==FIXED){
            data->fixed_cov->solve(tmpvec,tmpvec2);
          }
          stat_cov->Ltmult(tmpvec2,oldgrad);
          int flag=0;
          for(i=0;i<subgroup_size;i++){
            oldgrad[i]=sigma->value*oldgrad[i]-theta->value[i];
            if(oldgrad[i]>Hsub){
              flag=1;
              oldgrad[i]=Hsub;
            }
            if(oldgrad[i]<(-Hsub)){
              oldgrad[i]=-Hsub;
              flag=1;
            }
          }
          if(flag){
            //cout << "Maximum current gradient exceeded.\n";
          }
          theta->propose(theta->prop_sd_subgroup/sigma->value,subgroup_size);  // random noise
          stat_cov->Lcheck(theta->value,theta->value_stored);
          
          for(i=0;i<subgroup_size;i++){  // here's the proposal mean
            theta->value[i]+=(theta->prop_sd*theta->prop_sd)*oldgrad[i]/(2.0*sigma->value);
          }
          
          theta_to_ksi();
          ksi_to_psi();
          
          // gradient at proposed value for use in Hastings ratio
          if(which==ALPHA){
            for(i=0;i<subgroup_size;i++){
              tmpvec[i]=data->sum_data[i]-data->num_train_years*psi->value[i];
            }
          }
          else{
            for(i=0;i<subgroup_size;i++){
              tmpvec[i]=data->sum_data_by_t[i]-data->sum_t_squ*psi->value[i];
            }
          }
          data->stat_cov->solve(tmpvec,tmpvec2);
          stat_cov->Ltmult(tmpvec2,newgrad);
          flag=0;
          for(i=0;i<subgroup_size;i++){
            newgrad[i]=sigma->value*newgrad[i]-theta->value[i];
            if(newgrad[i]>Hsub){
              newgrad[i]=Hsub;
              flag=1;
            }
            if(newgrad[i]<(-Hsub)){
              newgrad[i]=-Hsub;
              flag=1;
            }
          }
          if(flag){
            //cout << "Maximum new gradient exceeded\n";
          }
          logprior_diff=theta->evaluate_prior();
          
          for(i=0;i<subgroup_size;i++){
            tmpvec[i]=sigma->value*(theta->value_stored[i]-theta->value[i])-(theta->prop_sd*theta->prop_sd/2.0)*newgrad[i];
            tmpvec2[i]=sigma->value*(theta->value[i]-theta->value_stored[i])-(theta->prop_sd*theta->prop_sd/2.0)*oldgrad[i];
            logprior_diff+=-0.5*(tmpvec[i]*tmpvec[i]-tmpvec2[i]*tmpvec2[i])/(theta->prop_sd*theta->prop_sd);
          }
          
          if(which==ETA){
            data->datacov_change(which,NOT_FIRST);
          }
          
          accepted=accept_lik(logprior_diff);
          
          theta->update(accepted);
          ksi->update(accepted);
          psi->update(accepted);
          
          if(basic_mode==POSTMEAN && accepted){
            //postmean_update(accepted);
            ksi_propose(TRUE);
          }
          
          data->update(accepted,which);  // update likelihood and data covariance matrix  -- but only for function
          num_accept_theta_subgroup+=accepted;
          run_num_accept_theta_subgroup+=accepted;
        }
      }
    }
  }
}

////////////////////////////////////////
// Class hyperparm_vec  
////////////////////////////////////////

hyperparm_vec::hyperparm_vec(int size_in,double* init,double prop_sd_in,double prop_sd_subgroup_in){
  prop_sd=prop_sd_in;   
  prop_sd_subgroup=prop_sd_subgroup_in;
  if(!prop_sd){
    fixed=TRUE;
  }
  else{
    fixed=FALSE;
  }
  if(prop_sd_subgroup){
    subgroup_fixed=FALSE;
  }
  else{
    subgroup_fixed=TRUE;
  }
  size=size_in;
  value=init;             // value and init share storage - CAREFUL
  value_stored=new double[size];
}

// initialization for parameter that is a function/transformation of another parameter sampled during MCMC
hyperparm_vec::hyperparm_vec(int size_in){
  size=size_in;                                    
  value=new double[size];
  value_stored=new double[size];
}

hyperparm_vec::hyperparm_vec(int size_in,double prop_sd_in){
  prop_sd=prop_sd_in;   
  size=size_in;                                    
  value=new double[size];
  value_stored=new double[size];
}

hyperparm_vec::~hyperparm_vec(void)  {
  delete [] value;
  delete [] value_stored;
}

void hyperparm_vec::propose(double alternate_prop_sd){
  double tmp=prop_sd;
  if(alternate_prop_sd>0.0){
    tmp=alternate_prop_sd;
  }
  for(int i=0;i<size;i++){
    value[i]+=tmp*gasdev(idum);  // normal proposal 
  }
}

void hyperparm_vec::propose(double alternate_prop_sd,int subgroup_size){
  double tmp=prop_sd_subgroup;
  if(alternate_prop_sd>0.0){
    tmp=alternate_prop_sd;
  }
  for(int i=0;i<subgroup_size;i++){
    value[i]+=tmp*gasdev(idum);  // normal proposal 
  }
}

double hyperparm_vec::evaluate_prior(void){          // MVN(0,1) prior
  logprior=mvn_dens(size,value,1.0);
  if(zero(logprior-DENSITY_ERROR)){
    cerr << "Warning: Possible error in log density for: " << value << endl;
  }
  return(logprior-logprior_stored);
}

void hyperparm_vec::update(int accepted)       
{
  if(accepted){                                // store newly-accepted values
    logprior_stored=logprior;
    copy(size,value,value_stored);
  }
  else{                                        // revert to stored values
    logprior=logprior_stored;
    copy(size,value_stored,value);
  }
}


////////////////////////////////////
// Class hyperparm_scal
////////////////////////////////////


hyperparm_scal::hyperparm_scal(double init,double parm1_in,double parm2_in,double prop_sd_in,prior_dens_fxn_ptr prior_dens_fxn_in,transform_fxn_ptr transform_in){
  transform=transform_in;
  prior_dens_fxn=prior_dens_fxn_in;
  prop_sd=prop_sd_in;
  if(!prop_sd){
    fixed=TRUE;
  }
  else{
    fixed=FALSE;
  }
  base_value=init;  
  value=transform(base_value,0);
  parm1=parm1_in;
  parm2=parm2_in;
}

double hyperparm_scal::propose(double alternate_prop_sd){                      // normal proposal density
  double dev=gasdev(idum);
  if(alternate_prop_sd>0.0){
    base_value+=alternate_prop_sd*dev;
  }
  else{
    base_value+=prop_sd*dev;
  }
  value=transform(base_value,0);
  return(dev);
}

void hyperparm_scal::propose(double alternate_prop_sd,double term){                      // normal proposal density
  if(alternate_prop_sd>0.0){
    base_value+=alternate_prop_sd*gasdev(idum);
  }
  else{
    base_value+=prop_sd*gasdev(idum);
  }
  base_value+=prop_sd*term;
  value=transform(base_value,0);
}

void hyperparm_scal::update(int accepted){
  if(accepted){                                      // store newly-accepted values
    logprior_stored=logprior;
    value_stored=value;
    base_value_stored=base_value;                    // irrelevant if no transform 
  }
  else{                                              // revert to stored values
    logprior=logprior_stored;
    value=value_stored;
    base_value=base_value_stored;
  }
}

double hyperparm_scal::evaluate_prior(){
  logprior=prior_dens_fxn(base_value,parm1,parm2);           // normal prior density (for transformed RV)
  if(zero(logprior-DENSITY_ERROR)){
    cerr << "Warning: Possible error in log density for: " << value << endl;
  }
  if(zero(logprior-NEG_INFTY)){
    return(logprior);
  }
  else{
    return(logprior-logprior_stored);
  }
}













